# Copyright 2016, 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt


"""
netq - Query data across all nodes in fabric

Usage:
   netq help [<text-keywords>]
   netq help verbose
   netq help list

Options:
   all                          : Information about all sub components
   around                       : Go back in time to around ...
   between                      : Specify start and end times for output desired
   config                       : Configure NetQ
   count                        : Count of matching entries
   example                      : Show examples of usage and workflow
   help                         : Show usage info
   ip                           : IPv4 related info
   ipv6                         : IPv6 related info
   json                         : Provide output in JSON
   list                         : Show all help commands
   notifier                     : Start/Stop/Restart/Query status of netq notifier
   origin                       : Owner of route or mac
   show                         : Show fabric-wide info about specified object
   verbose                      : More help on help
   <example>                    : name of example
   <hostname>                   : Name of the remote node you want to query
   <ip-server>                  : IP address of DB server
   <ipv4>                       : IPv4 address (no mask)
   <ipv6>                       : IPv6 address (no mask)
   <ip>                         : IPv4 or v6 address (no mask)
   <ipv4/prefixlen>             : IPv4 address with mask
   <ipv6/prefixlen>             : IPv6 address with mask
   <ip/prefixlen>               : IPv4 or IPv6 address with mask
   <mac>                        : MAC address
   <role>                       : Docker Swarm Node Role (worker, master)
   <text-time>                  : Time string such as 10m, 30s, 1h
   <text-endtime>               : Time string such as 10m, 30s, 1h
   <vrf>                        : Backend VRF
   <text-vlan>                  : VLAN

"""
from __future__ import absolute_import

import functools
import json
import os
import sys
from datetime import datetime
import syslog
import logging
import re
from glob import glob
import imp as importlib
import inspect
from functools import wraps
from operator import itemgetter

import tabulate
from network_docopt import (
    NetworkDocopt,
    get_network_docopt_info
)

from netq_lib.common import utils
from netq_lib.common.enums import (
    BgColors,
    DiffState,
)
from netq_apps.lib.netq import (
    NetQ,
)
from netq_apps.modules.core.netq.config import (
    conf_get_experimental,
    conf_get_addons
)
import netq_apps.lib.completers as completers
import netq_apps

_PROC_NAME = 'netq-cii'
args = None
RC_SUCCESS = 0
RC_CMD_NOT_FOUND = -2
MAX_RESULTS_PER_SHOW = 5000

class Context(object):
    '''Allow passage of some key params between lower function
    and decorator'''
    def __init__(self, cli):
        self.col_headers = []
        self.col_sizes = []
        self.sort_key = None
        self.reverse = False
        self.changes = True if cli.get('changes') else False
        self.count = True if cli.get('count') else False


class _MyTabulate(object):
    @staticmethod
    def __line_row(colsizes, *_):
        return ' '.join('-' * colsize for colsize in colsizes)

    @classmethod
    def __build_row(cls, colsizes, sep, cell_values, _, colaligns):
        values_with_attrs = []
        new_cell_values = []
        for cell_value, colsize, colalign in zip(cell_values,
                                                 colsizes,
                                                 colaligns):
            color = None
            if colalign == 'right':
                fmt = '{0:>%d}' % colsize
            else:
                fmt = '{0:<%d}' % colsize
            key = cell_value.strip()
            if key.startswith(BgColors.RED):
                color = BgColors.RED
                key = fmt.format(utils.strip_colors(key).strip())
            elif key.startswith(BgColors.YELLOW):
                color = BgColors.YELLOW
                key = fmt.format(utils.strip_colors(key).strip())
            elif key.startswith(BgColors.GREEN):
                color = BgColors.GREEN
                key = fmt.format(utils.strip_colors(key).strip())
            elif key.startswith(BgColors.GRAY):
                color = BgColors.GRAY
                key = fmt.format(utils.strip_colors(key).strip())
            else:
                key = fmt.format(key)
            if len(key) > colsize:
                idx = key.rfind(sep, 0, colsize)
                if idx == -1:
                    idx = colsize - 1
                if color:
                    new_cell_values.append(utils.color_wrap(color,
                                                            key[idx + 1:]))
                else:
                    new_cell_values.append(key[idx + 1:])
                key = key[:idx + 1] + ' ' * (colsize - idx - 1)
            else:
                new_cell_values.append('')
            if color:
                values_with_attrs.append(utils.color_wrap(color, key))
            else:
                values_with_attrs.append(key)
        line = ' '.join(values_with_attrs).rstrip()
        if any(new_cell_values):
            return line + '\n' + cls.__build_row(colsizes, sep,
                                                 new_cell_values, _,
                                                 colaligns)
        else:
            return line

    @classmethod
    def paginate(cls, data, headers, colsizes, sep=', ', first=False):
        colsizes = [max(x, len(y)) for x, y in zip(colsizes, headers)]
        linebelowheader = functools.partial(cls.__line_row, colsizes)
        headerrow = datarow = functools.partial(cls.__build_row, colsizes, sep)
        simple = tabulate.TableFormat(lineabove=None,
                                      linebelowheader=linebelowheader,
                                      linebetweenrows=None,
                                      linebelow=None,
                                      headerrow=headerrow,
                                      datarow=datarow,
                                      padding=0,
                                      with_header_hide=['lineabove',
                                                        'linebelow'])
        if first:
            return tabulate.tabulate(data, headers=headers, tablefmt=simple)
        else:
            return tabulate.tabulate(data, headers=[], tablefmt=simple)


def _build_diff(entries, state, mods):
    output = []
    for entry in entries:
        if entry is None:
            state += 1
            continue
        state_str = DiffState.MAP[state]
        if state == DiffState.MODIFIED:
            mods.append(entry)
            if len(mods) == 2:
                if mods[0][:-1] != mods[1][:-1]:
                    output.append((state_str,) + tuple(mods[0]))
                    output.append(('',) + tuple(mods[1]))
                mods[:] = []
        else:
            output.append((state_str,) + tuple(entry))
    return output, state


def netq_tabulate(output, headers=None, jsonify=False, first=True,
                  colsizes=None):
    if jsonify:
        out = []
        sanitized_headers = []
        json_out = []
        for header in headers:
            if header.startswith('#'):
                header = re.sub(r'#', 'num', header)
            sanitized_headers.append(
                ''.join(i.lower() if idx == 0 else i.lower().title()
                        for idx, i in enumerate(header.split(' ')))
            )
        for ele in output:
            out.append(dict((key, utils.strip_colors(item))
                            for key, item in zip(sanitized_headers, ele)))
        json_out.append(out)
        return json_out[0]
    else:
        # Stringify timestamps
        new_output = []
        for row in output:
            new_ele = []
            for ele in row:
                if isinstance(ele, float):
                    try:
                        datetime.fromtimestamp(ele)
                        new_ele.append(utils.pretty_date(ele))
                        continue
                    except Exception:  # pylint: disable=broad-except
                        pass
                new_ele.append(ele)
            new_output.append(new_ele)
        if colsizes is not None:
            p_out = _MyTabulate.paginate(new_output, headers=headers,
                                         colsizes=colsizes, first=first)
        else:
            p_out = tabulate.tabulate(new_output, headers)
        if first and p_out:
            print p_out
        elif p_out:
            print p_out[p_out.find('\n', p_out.find('\n') + 1) + 1:]

# These cli_show* routines never print the JSON output. This allows the
# JSON output to be aggregated. This is done already with 'show changes'
# command


class NetqCLI(NetworkDocopt):
    def __init__(self, daemon, logger):

        ''''''
        global __doc__

        self.callbacks = {}
        self.completers = []
        self.changes_cb = {}
        self._logger = logger

        # Make sure we update __doc__ as its used in help display
        self.cmdstrings = self.add_cmds(__doc__)
        NetworkDocopt.__init__(self, self.cmdstrings, daemon)

        for completer in self.completers:
            completer(self)

        # Add custom completers
        self.add_tag_matcher('<hostname>', completers.hostname_match, True)
        self.add_tab_complete_hook('<hostname>', completers.hostname_complete)

        self.short_doc_string = '''
    netq command is in one the of the following forms:
       netq config (config commands)
       netq [<hostname>] (check|show) <object> <options>
       netq trace <options>
       netq resolve
       netq (agent|notifier) (<start|stop|restart|status>) [vrf <vrf>]

    Type:
       'netq show agents' to see the list of hosts monitored by netq
       'netq help verbose' to get the help on help
       'netq help list' to dump all the commands and keyword help
       'netq example' to see examples of usage and possible workflow
       'netq help <text-keywords>' ('netq help check bgp', for eg.) to see
       commands matching keywords'''
        self.verbose_doc_string = '''
       netq commands have the following format:
           netq [<hostname>] action object [options]

       [] denotes an optional parameter or keyword
       <> denotes a parameter to be specified such as an IP prefix, addr etc.

       Hitting the TAB key will automatically show the available options.
       Partial keywords are also accepted; e.g.: 'netq show ip ro 4.0.0.1'.

       action is one of:
           check       :  Perform fabric-wide checks
           config      :  Configure NetQ
           example     :  Show examples of usage and workflow
           help        :  Show usage info
           resolve     :  Annotate input with names and interesting info
           show        :  Show fabric-wide info about specified object
           trace       :  Control plane trace path across fabric

       object is what the information is requested about such as interfaces,
       routes, services, etc.

       Options are used to further qualify the output desired such as
       asking for JSON output or requesting information about a specific object
       within the main object.

       Time based displays are governed by two options:
           around            : Go back in time to around ...
           between           : specify start and end times for output desired

       Time itself is specified as either a relative time such as 1s, 10m, 1d
       etc. or precise time '%Y-%m-%dT%H:%M:%S.%fZ' or 'd:h:m:s'.
        '''

    def add_cmds(self, doc):
        '''Construct complete command set and return for docopt to process'''
        usage_idx = doc.find('Usage:')
        options_idx = doc.find('Options:')
        usage = doc[:options_idx]
        post_usage = doc[options_idx:]

        mod_usage, mod_options, mod_callbacks = self._add_module_cmds()

        # new_usage = (usage + mod_usage + dyn_usage + post_usage +
        #             mod_options + dyn_options)

        new_usage = usage + mod_usage + post_usage + mod_options

        return new_usage

    def _add_module_cmds(self):
        options = ''
        usage = ''
        callbacks = {}
        matching_argspec = [['cli', 'netq'], ['cli', '_'], ['_', 'netq'],
                            ['_', '_']]
        main_dir = os.path.dirname(inspect.getfile(netq_apps))
        python_version = '%d.%d' % (sys.version_info.major,
                                    sys.version_info.minor)
        local_dir = '/usr/local/lib/%s/dist_packages/netq_apps/modules/' % (
            python_version)

        # Always parse in this order: core, experimental, addons under main
        # python dist-packages dir, followed by experimental and addons under
        # local lib
        dirstocheck = [main_dir + '/modules/core']
        load_exp = conf_get_experimental()
        load_addons = conf_get_addons()
        if load_exp:
            dirstocheck.append(main_dir + '/modules/experimental')
        if load_addons:
            dirstocheck.append(main_dir + '/modules/addons')

        # Now for the local dirstocheck
        dirstocheck.append(local_dir + '/modules/core')
        if load_exp:
            dirstocheck.append(local_dir + '/modules/experimental')
        if load_addons:
            dirstocheck.append(local_dir + '/modules/addons')

        for moddir in dirstocheck:
            if not os.path.exists(moddir):
                continue
            # traverse root directory, and list directories as dirs and
            # files as files
            for root, dirs, files in os.walk(moddir):
                for filename in sorted(files):
                    modname, extn = os.path.splitext(os.path.basename(filename))
                    mname = root + '/' + filename
                    try:
                        if extn == '.so':
                            mod = importlib.load_dynamic(modname, mname)
                        elif extn == '.py' and modname != '__init__':
                            mod = importlib.load_source(modname, mname)
                        else:
                            continue
                    except ImportError as ex:
                        self._logger.error('Unable to import module %s.'
                                           'Error is %s' % (mname, str(ex)))
                        continue
                    except Exception as ex:
                        self._logger.error('Unable to import module %s.'
                                           'Error is %s' % (mname, str(ex)))
                        continue
                    if hasattr(mod, 'build_doc_string'):
                        doc = mod.build_doc_string()
                    else:
                        doc = mod.__doc__
                    if not doc:
                        continue
                    usage_idx = doc.find('Usage:') + len('Usage:') + 1
                    options_idx = doc.find('Options:')
                    if usage_idx:
                        usage += doc[usage_idx:options_idx]
                    if options_idx:
                        options_idx += len('Options:') + 1
                        options += doc[options_idx:]
                    if (not hasattr(mod, 'app') or
                            not hasattr(mod.app, 'routes')):
                        continue
                    routes_from = mod.app.routes
                    if not isinstance(routes_from, dict):
                        self._logger.error(
                            'Unable to import module %s. Routes is not dict'
                            % (mname))
                        continue

                    for key, val in routes_from.iteritems():
                        if key in self.callbacks:
                            self._logger.warning(
                                'Duplicate route %s in module %s. '
                                'Prev route from %s' % (
                                    key, modname, self.callbacks[key]['module']))
                            continue

                        self.callbacks[key] = {'module': modname, 'cb': val}

                    # Add custom completers/matchers, if any
                    if (hasattr(mod, 'add_completers') and
                            callable(mod.add_completers)):
                        self.completers.append(mod.add_completers)

                    if (hasattr(mod, 'get_changes') and
                            callable(mod.get_changes)):
                        self.changes_cb[modname] = mod.get_changes

                    self._logger.info('Successfully imported module %s from %s'
                                      % (modname, mname))
        return usage, options, callbacks

    def process_cmd(self, cli, netq):
        # TBD: THIS NEEDS WORK. DON't CHECK IN THIS CODE LIKE THIS
        if self.args.get('<hostname>'):
            start_arg = 2
        else:
            start_arg = 1
        end_arg = start_arg + 4

        for x in range(end_arg, start_arg, -1):
            cbstr = ' '.join(str(a) for a in self.argv_expanded[start_arg:x])
            data = self.callbacks.get(cbstr)
            if data:
                cli['module_name'] = data['module']
                if cbstr == 'show changes':
                    # Show changes is special because it needs additional param
                    return data['cb'](cli, netq, self.changes_cb)
                else:
                    return data['cb'](cli, netq)

        return RC_CMD_NOT_FOUND

    def run(self, argv, uid, gid):
        """
        'parent' is the NetDaemon object used for checking user credentials.
        """
        global __doc__

        len_argv = len(argv)

        # Handle help separately
        # Special check for handling "netq help verbose" and "netq help legend"
        if len_argv == 1:
            print self.short_doc_string
            return 0
        elif argv[1] == 'help' or argv[-1] == 'help':
            if len_argv == 2:
                print self.short_doc_string
                return 0
            if argv[2] == 'verbose':
                print self.verbose_doc_string
                return 0
            if argv[2] == 'list':
                print self.cmdstrings
                return 0
            else:
                output = []
                print_section = False
                section = ''
                for line in self.cmdstrings.splitlines():
                    if not line:
                        continue
                    if line == 'Usage:':
                        section = 'Commands:'
                        print_section = True
                    elif line == 'Options:':
                        section = 'Keywords:'
                        print_section = True
                    prev_match = True
                    if argv[-1] == 'help':
                        for ele in argv[1:-1]:
                            if not re.search(r'\b%s\b' % ele, line):
                                prev_match = False
                                break
                    else:
                        for ele in argv[2:]:
                            if not re.search(r'\b%s\b' % ele, line):
                                prev_match = False
                                break
                    if prev_match:
                        if print_section:
                            print section
                            print_section = False
                        print line
            return 0

        NetworkDocopt.run(self, argv)  # Populates self.args
        args = self.args               # Convenience variable to avoid typing "self"

        start_time = end_time = None
        if args.get('around') or args.get('between'):
            start_time = args.get('<text-time>')
        if args.get('and'):
            end_time = args.get('<text-endtime>')
        jsonify = args.get('json')

        netq = NetQ(None, start_time, end_time)

        # if user is trying 'config add/del server' then
        # ignore empty server config
        if not (args.get('config') and args.get('server')):
            if netq.get_config_server_addr() is None:
                raise Exception("Backend server not specified")

        rc = RC_SUCCESS
        cli = args

        status = self.process_cmd(cli, netq)
        if status == RC_CMD_NOT_FOUND:
            if not jsonify:
                print 'Unable to find command %s' % ' '.join(argv)
            else:
                json.dumps({'error': 'Unable to find command %s' %
                            ' '.join(argv)})

        if cli.get('parser_reinit', False):
            self.callbacks = {}
            self.completers = []
            self.cmdstrings = self.add_cmds(__doc__)

            NetworkDocopt.__init__(self, self.cmdstrings, True)

            for completer in self.completers:
                completer(self)

            # Add custom completers
            self.add_tag_matcher('<hostname>', completers.hostname_match, True)
            self.add_tab_complete_hook('<hostname>',
                                       completers.hostname_complete)

            print 'Parser reloaded'
        return status


def netq_show(show_func):
    '''Decorator to wrap netq show functions'''

    @wraps(show_func)
    def __wrapper(cli, netq):

        context = Context(cli)
        first = True
        json_out = []
        output = []
        sort_key = None
        reverse = False
        entry_count = 0
        truncated_result = False

        jsonify = True if cli.get('json') else False
        module = cli.get('module_name', 'results')

        if context.count:
            out = show_func(cli, netq, context)
            if not (isinstance(out, int) or
                    isinstance(out, float) or
                    isinstance(out, long)):
                out = 0
            if not jsonify:
                print 'Count of matching {} records: {}'.format(module, out)
            else:
                print json.dumps({'count': out})
            return RC_SUCCESS

        for entry in show_func(cli, netq, context):
            if first:
                if not jsonify:
                    print 'Matching {} records are:'.format(module)
                # Build the column headers if not specified
                if not context.col_headers:
                    context.col_headers = [e for e in entry.keys() if e != 'error']
                first = False

            vals = process_show_entry(netq, entry)
            if vals:
                output.append(vals)
                entry_count += 1
                
            if entry_count > MAX_RESULTS_PER_SHOW:
                if jsonify:
                    print 'Showing only first {} entries. Filter for more '\
                        'complete result.'.format(MAX_RESULTS_PER_SHOW)
                truncated_result = False
                break

        if not output:
            if not jsonify:
                if context.changes:
                    print 'No changes in {} found'.format(module)
                else:
                    print 'No matching {} records found'.format(module)

        if context.sort_key:
            sort_key = context.sort_key
            reverse = context.reverse

        if not sort_key:
            if context.changes:
                sort_key = itemgetter(-1)
                reverse = True
            else:
                sort_key = itemgetter(0)

        if output:
            if context.col_sizes:
                json_out = netq_tabulate(sorted(output, key=sort_key,
                                                reverse=reverse),
                                         headers=context.col_headers,
                                         colsizes=context.col_sizes,
                                         jsonify=jsonify)
            else:
                json_out = netq_tabulate(sorted(output, key=sort_key,
                                                reverse=reverse),
                                         headers=context.col_headers,
                                         jsonify=jsonify)
        if jsonify:
            print json.dumps({module: json_out,
                              'truncatedResult': truncated_result}, indent=4)

        return RC_SUCCESS

    return __wrapper


def process_show_entry(netq, entry):
    '''Process a single entry and return the list'''
    vals = []
    if 'error' in entry:
        color = BgColors.RED
    elif (('hostname' in entry and
           entry['hostname'] in netq.rotten_nodes) or
          ('Hostname' in entry and
           entry['Hostname'] in netq.rotten_nodes)):
        color = BgColors.GRAY
    else:
        color = BgColors.ENDC
    for key in entry:
        if key == 'error':
            # this key is only used to set the color of the text
            continue
        if type(entry[key]) == str:
            vals.append(utils.color_wrap(color, entry[key]))
        else:
            vals.append(entry[key])
    return vals
