# Copyright 2016, 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

#'''
#Telemetry Service Configuration
#
#Usage:
#   netq config ts ha prepare <ip-host> <ip-master> <wildcard-replicas>
#   netq config ts ha finish
#   netq config ts ha restart
#   netq config ts cluster status [<ip-host>]
#   netq config ts cluster reset [<ip-master> <wildcard-replicas>]
#   netq config ts cluster failover [<ip-host>]
#   netq config ts cluster quorum [<ip-host>]
#
#Options:
#   prepare                      : Begin HA configuration
#   finish                       : End HA configuration
#   restart                      : Start or restart
#   status                       : Status
#   update                       : Update configuration
#   reset                        : Reset configuration
#   host-ip                      : Host to check or perform actions on
#   quorum                       : Force cluster quorum check
#   failover                     : Force cluster failover
#   ts                           : Telemetry Server
#   ha                           : High Availability
#   cluster                      : Cluster management
#   services                     : TS services
#   <ip-host>                    : Telemetry Server IP address
#   <ip-master>                  : Master data node for cluster
#   <wildcard-replicas>          : Replica data nodes for cluster
#   <wildcard-servers>           : One or more of the cluster nodes
#   full                         : Return complete systemd command response from systemctl
#'''

from __future__ import absolute_import
from __future__ import print_function

import os
import time
from functools import wraps

from netq_apps.modules import NetqModule, RC_SUCCESS, RC_FAIL
from netq_apps.modules.core.netq.backend.config_ha_impl import (
    Server,
    get_redis_host,
    restart_cluster,
    update_cluster_configuration,
    update_runtime_cluster,
    check_current_cluster_master,
    request_quorum,
    check_replication,
    check_sentinel,
    check_sentinels,
    force_cluster_failover,
    reset_redis_sentinels,
    status_services,
    get_settings,
    is_ts_server,
    restart_local_netqd,
)

from netq_apps.modules.core.netq.backend.config_ha_tools import rc_fail, get_wildcard_ips

__CLUSTER = 'cumulus-netq'
__REDIS_P = 6379
__SENTL_P = 26379


### TOOLS


def _process_response_rec(rec):
    if isinstance(rec, tuple):
        rc, msg = rec
        print(msg)

    if isinstance(rec, basestring):
        print(rec)


def _process_response(resp,timeout=0):
    if resp:
        if isinstance(resp, list):
            for rec in resp:
                if isinstance(rec,list):
                    _process_response(rec)
                else:
                    _process_response_rec(rec)
        else:
            _process_response_rec(resp)

    timeout = 0 if timeout < 0 else (timeout if timeout < 30 else 30)
#    if timeout:
#        time.sleep(timeout)

    return RC_SUCCESS


class SettingsException(Exception):
    pass

def host_ip(cli):
    host = get_settings(cli).get('<ip-host>')
    if host:
        return host
    raise SettingsException('Please provide a ip-host')


def master_ip(cli):
    master = get_settings(cli).get('<ip-master>')
    if master:
        return master
    raise SettingsException('Please provide a ip-master')


def servers_ips(cli):
    rc,ips = get_wildcard_ips(get_settings(cli).get('<wildcard-servers>'))
    if rc == RC_FAIL:
        raise SettingsException('Please provide wildcard-servers')
    return (rc,ips)

def replicas_ips(cli):
    rc,ips = get_wildcard_ips(get_settings(cli).get('<wildcard-replicas>'))
    if rc == RC_FAIL:
        raise SettingsException('Please provide wildcard-replicas')
    return (rc,ips)

def ensure_environ(f):
    @wraps(f)
    def check_and_load_environ(*args, **kwargs):
        # print(f.__name__)
        os.environ.update(
            {'APPLIANCE': '/etc/cts',
             'CFG_DIR': '/etc/cts',
             'DATA_DIR': '/var/data',
             'LOG_DIR': '/var/log/docker',
             'NETQ_GROUP_NAME': 'cumulus-netq',
             }
        )
        return f(*args, **kwargs)

    return check_and_load_environ

def on_ts_server(f):
    @wraps(f)
    def check_on_ts_server(*args, **kwargs):
        if not is_ts_server():
            return _process_response(rc_fail('Must be on Telemetry Server to configure.'))
        return f(*args, **kwargs)

    return check_on_ts_server


####### COMMANDS

app = NetqModule()


###### REDIS #####


@app.route('config ts cluster failover')
@ensure_environ
def cli_config_ts_cluster_failover(cli, netq):
    ''' Failover forces sentinel master election. Can be run from any server.'''
    host = Server(host_ip(cli), __SENTL_P)
    _process_response(force_cluster_failover(host, __CLUSTER))
    return _process_response(check_current_cluster_master(host,__CLUSTER))

@app.route('config ts cluster quorum')
@ensure_environ
def cli_config_ts_cluster_quorum(cli, netq):
    ''' Failover forces sentinel master election. Can be run from any server.'''
    host = Server(host_ip(cli), __SENTL_P)
    _process_response(request_quorum(host, __CLUSTER))
    return _process_response(check_current_cluster_master(host,__CLUSTER))

@app.route('config ts cluster reset')
@ensure_environ
def cli_config_ts_cluster_reset(cli, netq):
    '''Reset asks sentinels to forget inactive servers. Can be run from any server.'''

    master = Server(master_ip(cli), __REDIS_P)
    rc, ips = replicas_ips(cli)

    sentinels = [Server(ip, __SENTL_P) for ip in ips]
    _process_response(reset_redis_sentinels(__CLUSTER, master,[master]+sentinels))

    return cli_config_ts_ha_finish(cli,None)


@app.route('config ts cluster status')
@ensure_environ
def cli_config_ts_cluster_status(cli, netq):
    '''Check the current status of a cluster node's quorum, sentinel, replication. Can be run from any server.'''
    host = host_ip(cli)

    ret = []
    ret.append("Local HOST IP")
    ret.append(get_redis_host())
    ret.append(check_sentinel(Server(host, __SENTL_P), __CLUSTER))
    ret.append(check_replication(Server(host, __REDIS_P)))


    '''Check the status of the sentinels on the servers provided. Can be run from any server.'''
    rc, ips = servers_ips(cli)
    if rc == RC_FAIL:
        ret.append(rc_fail('Failed to find servers %s.' % ips))
        return _process_response(ret)

    sentinels = [Server(ip, __SENTL_P) for ip in ips]
    ret.append(check_sentinels(sentinels, __CLUSTER))

    return _process_response(ret)

@app.route('config ts ha restart')
@ensure_environ
@on_ts_server
def cli_config_ts_ha(cli, netq):
    '''Configures base TS into part of a HA TS cluster. Must be run on TS server.'''
    rc = RC_SUCCESS
    if cli.get('restart'):
        resp = ['Restarting Telemetry services.']
        resp.append(restart_cluster())
        return _process_response(resp)

@app.route('config ts ha prepare')
@ensure_environ
@on_ts_server
def cli_config_ts_ha(cli, netq):

    host = host_ip(cli)

    # TODO: Allow for non-default ports
    rc, replicas_or_msg = replicas_ips(cli)
    if rc == RC_FAIL:
        return _process_response(rc_fail(replicas_or_msg))

    # TODO: Allow for non-default ports
    master = master_ip(cli)
    if master in replicas_or_msg:
        return _process_response(
            rc_fail('Master node %s cannot also be a replica %s' % (master, ','.join(replicas_or_msg))))

    #TODO: An awesome use of a side-effect
    servers = ' '.join([master] + replicas_or_msg)
    servers_ips({'<wildcard-servers>':servers})

    resp = []
    resp.extend(update_cluster_configuration(host, __CLUSTER, master, replicas_or_msg))
    resp.append(restart_cluster())

    _process_response(resp)

    resp = ["Restart local NetQ CLI"]
    resp.append(restart_local_netqd())
    return _process_response(resp)


@app.route('config ts ha finish')
@ensure_environ
@on_ts_server
def cli_config_ts_ha_finish(cli, netq):

    host = host_ip(cli)

    # TODO: Allow for non-default ports
    rc, replicas_or_msg = replicas_ips(cli)

    # TODO: Allow for non-default ports
    master = master_ip(cli)
    if master in replicas_or_msg:
        return _process_response(
            rc_fail('Master node %s cannot also be a replica %s' % (master, ','.join(replicas_or_msg))))

    return _process_response(update_runtime_cluster(host, __CLUSTER, master, replicas_or_msg))



@app.route('config ts services status')
@ensure_environ
@on_ts_server
def cli_config_ts_services(cli, netq):
    '''Checks status of TS specific services. Must be run on TS server.'''
    _process_response(status_services(full=cli.get('full')))


